"""
SDQC loading file
"""
import os
import warnings
import configparser
from ast import literal_eval
import pysd
from pysd.translation.utils import make_python_identifier, make_coord_dict

from .checks_dict import checks_dict as checks

# keep missing values of external objects
pysd.external.External.missing = "keep"


def load_data(file_name):
    """
    Loads the data.

    Parameters
    ----------
    file_name: str
        File name to load external data from. It must be a Vensim model file
        (.mdl) or python PySD compatible file (.py).

    Returns
    -------
    exts: list
        The list of the PySD External objects used in the model.

    """
    file_extension = os.path.splitext(file_name)[1]

    if file_extension == ".mdl":
        model = pysd.read_vensim(file_name, initialize=False)
    elif file_extension == ".py":
        model = pysd.load(file_name, initialize=False)
    else:
        raise ValueError(
            "Invalid file extension.\n"
            + "The file must be Python file or Vensim's mdl file"
            )

    # Add original Vensim name information to de objects
    exts = model._external_elements
    namespace = model.components._namespace
    for ext in exts:
        ext.original_name = None
        for element in namespace:
            if ext.py_name in ['_ext_data_' + namespace[element],
                               '_ext_constant_' + namespace[element],
                               '_ext_lookup_' + namespace[element],
                               namespace[element]]:
                # original name
                ext.original_name = element
                # short name
                ext.py_short_name = namespace[element]

    return exts


def load_from_dict(elements):
    """
    Loads external outpust from a elements dictionary.

    Parameters
    ----------
    elements: dict
        Dictionary of the elements. Check documentation #TODO add link

    Returns
    -------
    exts: list
        The list of PySD External objects used in the model.

    """
    namespace = {}  # namespace of external objects
    subscript_dict = {}  # subscript dictionary

    for element, data in elements.items():
        if data['type'] == 'SUBSCRIPT':
            # add subscripts to the dict
            subscript_dict[element] = data['values']
        elif data['type'] == 'EXTERNAL':
            # add external variables to the namespace
            make_python_identifier(element, namespace)

    added = {'dataseries': {}, 'constant': {}}

    for element in namespace:
        for excel in elements[element]['excel']:
            # retrieve excel information
            filename = excel.get('filename')
            sheet = excel.get('sheet')
            cell = excel.get('cell')
            # TODO errors for missing filename/sheet/cell
            root = excel.get('root')
            x_row_or_col = excel.get('x_row_or_cols')
            coords = make_coord_dict(
                excel.get('subs') or {}, subscript_dict, terse=False)

            if x_row_or_col:
                # convert series data to pysdexternal Lookup object
                if element in added['dataseries']:
                    # add table to existing object
                    added['dataseries'][element].add(
                        filename, sheet, x_row_or_col, cell, coords)
                else:
                    # create new object
                    obj = pysd.external.ExtLookup(
                        filename, sheet, x_row_or_col, cell, coords,
                        root, '_ext_lookup_' + namespace[element])
                    obj.original_name = element
                    obj.py_short_name = namespace[element]
                    added['dataseries'][element] = obj
            else:
                # convert constant data to pysdexternal Constant object
                if element in added['constant']:
                    # add table to existing object
                    added['constant'][element].add(
                        filename, sheet, cell, coords)
                else:
                    # create new object
                    obj = pysd.external.ExtConstant(
                        filename, sheet, cell, coords,
                        root, '_ext_constant_' + namespace[element])
                    obj.original_name = element
                    obj.py_short_name = namespace[element]
                    added['constant'][element] = obj

    return list(added['dataseries'].values())\
        + list(added['constant'].values())


def load_config(config_file):
    """
    Load default configuration and user configuration if any.

    Parameters
    ----------
    config_file: str
      Path to the user configuration file. If None, no file is loaded.

    Returns
    -------
    config_dict: dict
      Dictionary with the configuration retrieved from the config file

    """
    # dictionary of arguments to load from config file and they type
    config_args = {}
    for check in checks.values():
        config_args[check['flag']] = 'bool'
        if 'args' in check and check['args']:
            for arg, (type, arg_in_func) in check['args'].items():
                config_args[arg] = type

    # TODO let define groups
    # TODO enable checks by dimension/subscript range

    config = configparser.ConfigParser()
    path2lib = os.path.dirname(os.path.abspath(__file__))
    config.read(os.path.join(path2lib, 'default-conf.ini'))
    config_dict = {'all': {}, 'dataseries': {}, 'constant': {}}
    if config_file:
        if not os.path.isfile(config_file):
            raise FileNotFoundError(
                f"No such file or directory: '{config_file}'")
        # read user's config file
        config.read(config_file)
    for section in config.keys():
        sec = section.lower()
        if sec not in config_dict:
            config_dict[sec] = {}
        for key in config[section].keys():
            # convert default values to a dictionary
            if key not in config_args:
                warnings.warn(f"{key} defined in the config file is not"
                              " used in any check")
            elif config_args[key] == 'bool':
                # Boolean value (bool)
                config_dict[sec][key] = config.getboolean(section, key)
            elif config_args[key] == 'float':
                # Numeric value (float)
                config_dict[sec][key] = config.getfloat(section, key)
            elif config_args[key] == 'int':
                # Integer value (int)
                config_dict[sec][key] = config.getint(section, key)
            elif config_args[key] == 'str':
                # Type value (str)
                config_dict[sec][key] = config.get(section, key)
            elif config_args[key] == 'list (float)':
                # Range value (list)
                value = config.get(section, key)
                if value[0] == '[':
                    # literla evaluation of a list
                    config_dict[sec][key] = literal_eval(value)
                else:
                    value = [float(val) for val in value.split()]
                    if all([val.is_integer() for val in value]):
                        # if all are integers convert them to int class
                        value = [int(val) for val in value]

                    config_dict[sec][key] = value
            else:
                raise ValueError(
                    f"Invalid value type definde in checks_dict for {key}")

    return config_dict



"""
SDQC main file
"""
import warnings
import numpy as np
import pandas as pd

from .loading import load_data, load_config, load_from_dict
from .tools import _add_config_to_ext
from .checks_dict import checks_dict as checks


def check(source, config_file=None, output='dataframe',
          verbose=False):
    """
    Main function to run the check.

    Parameters
    ----------
    source: str or dict
        File name or dictionary to load external data from.
        If a file name is given, it must be a Vensim model file
        (.mdl) or python PySD compatible file (.py). If a dict is
        given it should be compatible with loading.load_from_dict
        function (see #TODO link to documentation).

    config_file: str (optional)
        Path to the config file.

    output: str (optional)
        Options...

    verbose: bool (optional)
        If True the output from all checks is saved. If False (default)
        only the output from not passed checks will be saved.

    Returns
    -------
    out:

    """

    if isinstance(source, str):
        # load external objects from .mdl/PySD model in Python
        exts = load_data(source)
    elif isinstance(source, dict):
        # load data from a dictionary
        exts = load_from_dict(source)
    else:
        raise ValueError("Source must be a file_name or a dictionary.")

    # load configuration
    config = load_config(config_file)

    # add config information to each external object
    _add_config_to_ext(exts, config)

    # iterate over External objects to run the checks
    out = []
    for ext in exts:
        if ext.check_config['_type'] == 'constant':
            check_constant(ext, out, verbose)
        elif ext.check_config['_type'] == 'dataseries':
            check_dataseries(ext, out, verbose)

    # TODO add report output
    if output == "dataframe":
        return pd.DataFrame(out)
    else:
        return out


def check_constant(ext, out, verbose):
    """
    Run the checks in a ExtConstant type object.

    Parameters
    ----------
    ext: pysd.external.ExtConstant
        The object to pass the check. I must have the py_short_name,
        original_name and config attributes, setted previously.
    out: list
        The list to append the results of the checks.
    verbose: bool
        If True the output from all checks is saved. If False only the
        output from not passed checks will be saved.

    Returns
    -------
    None

    """
    # TODO enable checks by dimension/subscript range

    info = {
       'py_name': ext.py_name,
       'py_short_name': ext.py_short_name,
       'original_name': ext.original_name,
       'file': ext.files,
       'sheet': ext.sheets,
       'transposed': None,
       'cell': ext.cells,
       'coords': ext.coordss
       }

    try:
        ext.initialize()
    except Exception as err:
        warnings.warn(
            err.args[0]
            + f"\n\nNot able to initialize the following object:\n{info}"
            + "\nFind error details above.\n")
        return

    if ext.coordss[0]:
        constant = ext().values
    else:
        # if constant value is 0-D (float) no test can be passed
        return

    # Start checks
    for chk in checks:
        # Is check activated and for constant?
        if 'constant' in checks[chk]['targets']\
          and ext.check_config[checks[chk]['flag']]:
            new_info = _run_check([constant], chk, ext.check_config,
                                  'constant', info, verbose)
            if new_info:
                out.append(new_info)


def check_dataseries(ext, out, verbose):
    """
    Run the checks in a ExtData or ExtLookup type object.

    Parameters
    ----------
    ext: pysd.external.ExtData or pysd.external.ExtLookup
        The object to pass the check. I must have the py_short_name,
        original_name and config attributes, setted previously.
    out: list
        The list to append the results of the checks.
    verbose: bool
        If True the output from all checks is saved. If False only the
        output from not passed checks will be saved.

    Returns
    -------
    None

    """
    # TODO enable checks by dimension/subscript range

    info = {
       'py_name': ext.py_name,
       'py_short_name': ext.py_short_name,
       'original_name': ext.original_name,
       'file': ext.files,
       'sheet': ext.sheets,
       'transposed': None,
       'cell': ext.cells,
       'coords': ext.coordss
       }

    if hasattr(ext, 'time_row_or_cols'):
        # ExtLookup
        info['x_row_or_coll'] = ext.time_row_or_cols
    else:
        # ExtData
        info['x_row_or_coll'] = ext.x_row_or_cols

    try:
        ext.initialize()
    except Exception as err:
        warnings.warn(
            err.args[0]
            + f"\n\nNot able to initialize the following object:\n{info}"
            + "\nFind error details above.\n")
        return

    data = ext.data.values
    series = ext.data.coords[ext.data.dims[0]].values

    # Always check first for missing values in the series as we need
    # to remove them before starting
    new_info = info.copy()
    new_info['check_name'] = 'missing_values_series'
    new_info['check_description'] = checks['missing_values']['descr']
    new_info['check_target'] = 'series'
    new_info['check_arg'] = {}
    new_info['check_pass'], new_info['check_out'] =\
        checks['missing_values']['func'](series)

    # Remove missing values from series
    data = data[~np.isnan(series)]
    series = series[~np.isnan(series)]
    if verbose or not new_info['check_pass']:
        out.append(new_info)

    # Start checks
    for chk in checks:
        # Is check activated?
        new_info = None
        if ext.check_config[checks[chk]['flag']]:
            if 'series' in checks[chk]['targets']:
                # Check for series
                new_info = _run_check([series], chk, ext.check_config,
                                      'series', info, verbose)
            elif 'data' in checks[chk]['targets']:
                # Check for data
                new_info = _run_check([data], chk, ext.check_config,
                                      'data', info, verbose)
            elif 'dataseries' in checks[chk]['targets']:
                # Check for data + series
                new_info = _run_check([data, series], chk, ext.check_config,
                                      'dataseries', info, verbose)
            if new_info:
                out.append(new_info)


def _run_check(args, check_name, econfig, target, info, verbose):
    """
    Run a check with the given arguments.

    Parameters
    ----------
    args: list
        List of arguments to run a check, could be data, series,
        constant or data and series.
    check_name: str
        Name of the check to be passed.
    econfig: dict
        Configuration dictionary of the object. Needed to find extra
        arguments for the function.
    target: str
        The target of the check, must match the arguments passed in args.
    info: dict
        General info for the current reading block.
    verbose: bool
        If True the output from all checks is saved. If False only the
        output from not passed checks will be saved.

    Returns
    -------
    None

    """
    # create a new dictionary to sabe
    new_info = info.copy()

    # save name, description and target
    new_info['check_name'] = check_name
    new_info['check_description'] = checks[check_name]['descr']
    new_info['check_target'] = target

    if 'args' in checks[check_name] and checks[check_name]['args']:
        # read the arguments and save the info
        new_info['check_arg'] = {
            checks[check_name]['args'][arg][1]: econfig[arg]
            for arg in checks[check_name]['args']}
    else:
        # no arguments
        new_info['check_arg'] = {}

    # run the check
    new_info['check_pass'], new_info['check_out'] =\
        checks[check_name]['func'](*args, **new_info['check_arg'])

    # save the result if the check has not passed or verbose is True
    if not new_info['check_pass'] or verbose:
        return new_info



"""
SDQC checks file
"""
import numpy as np


def missing_values(data):
    """
    Checks for missing values on the data.

    Parameters
    ----------
    data: ndarray
        The data to check.

    Returns
    -------
    pass, outs: bool, list
        pass is True if no missing values have been found and False if
        missing values have been found. outs is a list with the number
        of missing values and the total number of values.

    """
    n_nan = int(np.sum(np.isnan(data)))
    n_values = int(np.prod(data.shape))

    if n_nan == 0:
        return True, [0, n_values]

    return False, [n_nan, n_values]


def missing_values_data(data, series, completeness):
    """
    Checks for missing values on the data giving the series.

    Parameters
    ----------
    data: ndarray
        The data to check.
    series: ndarray (1D)
        The series to check.
    completeness: str (optional)
        If completeness is 'any' (Default) the check will not pass
        is there is any missing value for a given series value. If
        completeness is 'all' the check will not pass all the data
        values are missing for a given series value (column).
        It only has effects when data is a matrix (2 or more dimensions).

    Returns
    -------
    pass, outs: bool, list
        pass is True if no missing values have been found and False if
        missing values have been found. outs is a list with a fisrt list
        of the number of missing values and the total number of values
        checked, and a second list with the series values for which the
        missing values where found. Note that the number of missing
        values and the total values depens on the completeness.

    """
    data_nan = np.isnan(data)
    n_nan = int(np.sum(data_nan))
    n_values = int(np.prod(data_nan.shape))

    if len(data.shape) >= 2:
        if completeness == 'all':
            data_nan = np.all(data_nan, axis=tuple(range(1, data_nan.ndim)))
            n_nan = int(np.sum(data_nan))
            n_values = int(np.prod(data_nan.shape))
        elif completeness == 'any':
            data_nan = np.any(data_nan, axis=tuple(range(1, data_nan.ndim)))
        else:
            raise ValueError("Invalid value for completeness argument.")

    if not any(data_nan):
        return True, [[0, n_values], []]

    missing = list(series[data_nan])
    return False, [[n_nan, n_values], missing]


def series_monotony(series):
    """
    Checks if series is monotonous.

    Parameters
    ----------
    series: ndarray (1D)
        The series to check.

    Returns
    -------
    pass, outs: bool, list
        pass is True if series is monotonous and False if not.
        outs is a list with the series range (minimum and maximum).

    """
    if all(np.diff(series) < 0) or all(np.diff(series) > 0):
        return True, [series[0], series[-1]]

    return False, [np.min(series), np.max(series)]


def series_range(series, srange):
    """
    Checks if series range is inside a given range.

    Parameters
    ----------
    series: ndarray (1D)
        The series to check.

    srange: list (len 2)
        The minimum and maximum value of the series.

    Returns
    -------
    pass, outs: bool, list
        pass is True if series is monotonous and False if not.
        outs is a list with the series range (minimum and maximum)
        and a list with the give srange.

    """
    smin, smax = np.min(series), np.max(series)
    if smin < srange[0] or smax > srange[1]:
        return False, [[smin, smax], srange]

    return True, [[smin, smax], srange]


def series_increment_type(series, type):
    """
    Checks if series range is inside a given range.

    Parameters
    ----------
    series: ndarray (1D)
        The series to check.

    type: str
        The series distribution ('linear')

    Returns
    -------
    pass, outs: bool, list
        pass is True if series is monotonous and False if not.
        outs is a list with the series step or the full series.

    """
    # TODO include more distributions ('exponential', 'logarithmic'...)
    if type == 'linear':
        ds = np.diff(series)
        if np.all(ds == ds[0]):
            # return True with the step
            return True, [ds[0]]
    else:
        raise ValueError(
            f"The argument {type} is not supported by series_increment_type.")

    return False, list(series)


def outlier_values(data, method, **kwargs):
    """
    Checks if there are possible outliers in the data using a
    multiple of the standard deviation.

    Parameters
    ----------
    data: ndarray
        The data to check.

    method: str
         The method to be used. Can be 'std' for standard deviation method
         or 'iqr' for interquartile range method.

    **kwargs: see below

    Keyword Args
    ------------
    nstd: float
        For 'std' method, the number of standard deviations
        to define outliers.

    niqr: float
        For 'iqr' method, the number of interquartile ranges
        to define outliers.

    Returns
    -------
    pass, outs: bool, list
        pass is True if series is monotonous and False if not.
        outs is a list with the number of possible outliers and
        the total number of values.

    """

    n_values = int(np.prod(data.shape))

    if method == 'std':
        # standard deviation method
        nstd = kwargs.get('nstd', 2)
        mean = np.nanmean(data)
        std = np.nanstd(data)
        n_outliers = int(np.sum(np.logical_or(data > mean + nstd*std,
                                              data < mean - nstd*std)))
    elif method == 'iqr':
        # interquartile range method
        niqr = kwargs.get('niqr', 1.5)
        q_low, q_up = np.percentile(data, 25), np.percentile(data, 75)
        iqr = q_up - q_low
        n_outliers = int(np.sum(np.logical_or(data > q_up + niqr*iqr,
                                              data < q_low - niqr*iqr)))
    else:
        raise ValueError(
            f"The method {method} is not supported by outlier_values."
        )

    if n_outliers == 0:
        return True, [0, n_values]

    return False, [n_outliers, n_values]



from setuptools import setup, find_packages

exec(open('sdqc/_version.py').read())

setup(
    name='sdqc',
    version=__version__,
    python_requires='>=3.7',
    author='Eneko Martin Martinez',
    author_email='eneko.martin.martinez@gmail.com',
    packages=find_packages(exclude=['docs', 'tests', 'dist', 'build']),
    url='https://gitlab.com/eneko.martin.martinez/sdqc',
    license = 'MIT',
    description='System Dynamics Data Quality Check with Python',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords=['System Dynamics', 'Vensim', 'Quality Check'],
    classifiers=[
    'Development Status :: 3 - Alpha',

    'Topic :: Scientific/Engineering :: Information Analysis',
    'Intended Audience :: Science/Research',

    'License :: OSI Approved :: MIT License',

    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    ],
    install_requires=[
        'pysd'
    ],
    include_package_data=True
)
